#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  base.py
"""
Base Classes
"""
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#  Copyright (c) 2015 Warren Smith
#  Copyright (c) Marco Sulla
#  Copyright (c) 2012 Santiago Lezica
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#


# stdlib
from abc import ABC, abstractmethod
from collections.abc import Mapping, MutableMapping

# 3rd party
from domdf_python_tools.doctools import is_documented_by


class DictWrapper(ABC):
	"""
	Absrtract Mixin class for classes that wrap a dict object or similar
	"""
	
	def __getitem__(self, key):
		return self._dict[key]
	
	def __contains__(self, key):
		return key in self._dict
	
	def __iter__(self):
		return iter(self._dict)
	
	def __len__(self):
		return len(self._dict)
	
	def __repr__(self):
		return f'<{self.__class__.__name__} {self._dict!r}>'

	@abstractmethod
	def copy(self, *args, **kwargs):
		pass

	def __copy__(self, *args, **kwargs):
		return self.copy()


class FrozenBase(DictWrapper, Mapping):
	"""
	Abstract Base Class for Frozen dictionaries
	
	Used by frozendict and FrozenOrderedDict.
	
	Custom subclasses must implement at a minimum ``__init__``,
	``copy``, ``fromkeys``.
	"""
	
	dict_cls = None
	
	@abstractmethod
	def __init__(self, *args, **kwargs):
		self._dict = self.dict_cls(*args, **kwargs)
		self._hash = None

	@classmethod
	@is_documented_by(dict.fromkeys)
	def fromkeys(cls, *args, **kwargs):
		return cls(dict.fromkeys(*args, **kwargs))


class MutableBase(DictWrapper, MutableMapping):
	"""
	Abstract Base Class for mutable dictionaries
	
	Used by NonelessDict and NonelessOrderedDict.
	
	Custom subclasses must implement at a minimum ``__init__``,
	``copy``, ``fromkeys``.
	"""
	
	dict_cls = None
	
	@abstractmethod
	def __init__(self, *args, **kwargs):
		self._dict = self.dict_cls(*args, **kwargs)
		self._hash = None
	
	def __setitem__(self, key, value):
		if value:
			self._dict[key] = value
	
	def __delitem__(self, key):
		del self._dict[key]
	
	@classmethod
	@is_documented_by(dict.fromkeys)
	def fromkeys(cls, *args, **kwargs):
		return cls(dict.fromkeys(*args, **kwargs))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  __init__.py
"""
Provides several useful custom dictionaries
"""
#
#  Copyright (c) 2019-2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#  Copyright (c) 2015 Warren Smith
#  Copyright (c) Marco Sulla
#  Copyright (c) 2012 Santiago Lezica
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#


from .alphadict import alphabetical_dict, AlphaDict
from .base import FrozenBase, MutableBase
from .bdict import bdict
from .frozendict import frozendict
from .nonelessdict import NonelessDict, NonelessOrderedDict
from .frozenordereddict import FrozenOrderedDict

__author__ = "Dominic Davis-Foster"
__copyright__ = "2020 Dominic Davis-Foster"

__license__ = "LGPLv3+"
__version__ = "0.1.2"
__email__ = "dominic@davis-foster.co.uk"

__all__ = [
		"__version__",
		"alphabetical_dict",
		"AlphaDict",
		"bdict",
		"FrozenBase",
		"frozendict",
		"FrozenOrderedDict",
		"MutableBase",
		"NonelessDict",
		"NonelessOrderedDict",
		]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  alphadict.py
"""
Provides AlphaDict, a frozen OrderedDict where the keys are stored alphabetically
"""
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#


# stdlib
from collections import OrderedDict

from .frozenordereddict import FrozenOrderedDict


def alphabetical_dict(**kwargs):
	return OrderedDict(sorted(kwargs.items()))


class AlphaDict(FrozenOrderedDict):
	def __init__(self, seq=None, **kwargs):
		"""
		Initialize an immutable, Alphabetised dictionary.
		The signature is the same as regular dictionaries.

		dict() -> new empty AlphaDict
		dict(mapping) -> new AlphaDict initialized from a mapping object's
			(key, value) pairs
		dict(iterable) -> new AlphaDict initialized as if via:
			d = {}
			for k, v in iterable:
				d[k] = v
		dict(**kwargs) -> new AlphaDict initialized with the name=value pairs
			in the keyword argument list.  For example:  dict(one=1, two=2)
		"""
		
		if seq:
			kwargs = dict(seq)
		
		print(sorted(kwargs.items()))
		super().__init__(sorted(kwargs.items()))




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  frozenordereddict.py
"""
Provides FrozenOrderedDict, an immutable ordered dictionary.
"""
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#  Copyright (c) 2015 Warren Smith
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#


# stdlib
import operator
from collections import OrderedDict
from functools import reduce

# this package
from .base import FrozenBase


class FrozenOrderedDict(FrozenBase):
    """
    An immutable OrderedDict.
    It can be used as a drop-in replacement for dictionaries where immutability is desired.
    """

    dict_cls = OrderedDict

    def __init__(self, *args, **kwargs):
        if hasattr(self, "_dict"):
            raise TypeError(f"`{self.__class__}` can only be initialised once.")

        super().__init__(*args, **kwargs)

    def copy(self, *args, **kwargs):
        new_dict = self._dict.copy()

        if args or kwargs:
            new_dict.update(OrderedDict(*args, **kwargs))

        return self.__class__(new_dict)

    def __hash__(self):
        if self._hash is None:
            self._hash = reduce(operator.xor, map(hash, self.items()), 0)

        return self._hash



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  frozendict.py
"""
Provides frozendict, a simple immutable dictionary.
"""
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#  Copyright (c) Marco Sulla
#  Copyright (c) 2012 Santiago Lezica
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#


# this package
from .base import FrozenBase


class frozendict(FrozenBase):
	"""
	An immutable wrapper around dictionaries that implements the complete
	:py:class:`collections.Mapping` interface. It can be used as a
	drop-in replacement for dictionaries where immutability is desired.
	"""
	dict_cls = dict
	
	def __init__(self, *args, **kwargs):
		if hasattr(self, "_dict"):
			raise TypeError(f"`{self.__class__}` can only be initialised once.")
		
		super().__init__(*args, **kwargs)
	
	def copy(self, **add_or_replace):
		return self.__class__(self, **add_or_replace)
	
	def __hash__(self):
		if self._hash is None:
			h = 0
			for key, value in self._dict.items():
				h ^= hash((key, value))
			self._hash = h
		return self._hash
	
	def sorted(self, *args, by="keys", **kwargs):
		"""
		Return a new `frozendict`, with the element insertion sorted.
		The signature is the same of builtin `sorted()` function, except for
		the additional parameter `by`, that is "keys" by default and can also
		be "values" and "items". So the resulting `frozendict` can be sorted
		by keys, values or items.

		If you want more complicated sorts, see the documentation of
		`sorted()`. Take into mind that the parameters passed to the `key`
		function are the keys of the `frozendict` if `by == "keys"`, and are
		the items otherwise.

		PS: Note that sort by keys and items are identical. The only
		difference is when you want to customize the sorting passing a custom
		`key` function. You *could* achive the same result using
		`by="values"`, since also sorting by values passes the items to the
		key function. But this is an implementation detail and you should not
		rely on it.
		"""
		
		if not self:
			return self
		
		sort_by_keys = by == "keys"
		sort_by_values = by == "values"
		
		if sort_by_keys:
			tosort = self.keys()
		elif sort_by_values:
			tosort = self.items()
		elif by == "items":
			tosort = self.items()
		else:
			raise ValueError(f"Unexpected value for parameter `by`: {by}")
		
		if sort_by_values:
			kwargs.setdefault("key", lambda item: item[1])
		
		it_sorted = sorted(tosort, *args, **kwargs)
		
		if it_sorted == list(tosort):
			return self
		
		if sort_by_keys:
			res = {k: self[k] for k in it_sorted}
		else:
			res = it_sorted
		
		return self.__class__(res)
	
	def __add__(self, other, *args, **kwargs):
		"""
		If you add a dict-like object, a new frozendict will be returned, equal
		to the old frozendict updated with the other object.
		"""
		
		tmp = dict(self)
		
		try:
			tmp.update(other)
		except Exception:
			raise TypeError(
					f"Unsupported operand type(s) for +: `{self.__class__.__name__}` and `{other.__class__.__name__}`") from None
		
		return self.__class__(tmp)
	
	def __sub__(self, other, *args, **kwargs):
		"""
		The method will create a new `frozendict`, result of the subtraction
		by `other`.

		If `other` is a `dict`-like, the result will have the items of the
		`frozendict` that are *not* in common with `other`.

		If `other` is another type of iterable, the result will have the
		items of `frozendict` without the keys that are in `other`.
		"""
		
		try:
			iter(other)
		except Exception:
			raise TypeError(
					f"Unsupported operand type(s) for -: `{self.__class__.__name__}` and `{other.__class__.__name__}`") from None
		
		try:
			res = {k: v for k, v in self.items() if (k, v) not in other.items()}
		except Exception:
			if not hasattr(other, "gi_running"):
				true_other = other
			else:
				true_other = tuple(other)
			
			res = {k: v for k, v in self.items() if k not in true_other}
		
		return self.__class__(res)
	
	def __and__(self, other, *args, **kwargs):
		"""
		Returns a new `frozendict`, that is the intersection between `self`
		and `other`.

		If `other` is a `dict`-like object, the intersection will contain
		only the *items* in common.

		If `other` is another iterable, the intersection will contain
		the items of `self` which keys are in `other`.

		Iterables of pairs are *not* managed differently. This is for
		consistency.

		Beware! The final order is dictated by the order of `other`. This
		allows the coder to change the order of the original `frozendict`.

		The last two behaviors breaks voluntarly the `dict.items()` API, for
		consistency and practical reasons.
		"""
		
		try:
			try:
				res = {k: v for k, v in other.items() if (k, v) in self.items()}
			except Exception:
				res = {k: self[k] for k in other if k in self}
		except Exception:
			raise TypeError(
					f"Unsupported operand type(s) for &: `{self.__class__.__name__}` and `{other.__class__.__name__}`") from None
		
		return self.__class__(res)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  bdict.py
"""
Provides bdict, a dictionary where keys and values are also stored the other way round
"""
#
#  Copyright (c) 2019-2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#


class bdict(dict):
	"""
	Returns a new dictionary initialized from an optional positional argument
	and a possibly empty set of keyword arguments.

	Each key:value pair is entered into the dictionary in both directions,
	so you can perform lookups with either the key or the value.

	If no positional argument is given, an empty dictionary is created.
	If a positional argument is given and it is a mapping object, a dictionary
	is created with the same key-value pairs as the mapping object.
	Otherwise, the positional argument must be an iterable object.
	Each item in the iterable must itself be an iterable with exactly two
	objects. The first object of each item becomes a key in the new
	dictionary, and the second object the corresponding value.

	If keyword arguments are given, the keyword arguments and their values are
	added to the dictionary created from the positional argument.

	If an attempt is made to add a key or value that already exists in the
	dictionary a ValueError will be raised

	Keys or values of "None", "True" and "False" will be stored internally as
	"_None" "_True" and "_False" respectively

	Based on https://stackoverflow.com/a/1063393 by https://stackoverflow.com/users/9493/brian
	"""
	
	def __init__(self, seq=None, **kwargs):
		super().__init__(self)
		if seq:
			for key, value in dict(seq).items():
				self.__setitem__(key, value)
		else:
			for key, value in kwargs.items():
				self.__setitem__(key, value)
	
	def __setitem__(self, key, val):
		if key in self or val in self:
			if key in self and self[key] != val:
				raise ValueError(f"The key '{key}' is already present in the dictionary")
			if val in self and self[val] != key:
				raise ValueError(f"The key '{val}' is already present in the dictionary")
		
		if key is None:
			key = "_None"
		if val is None:
			val = "_None"
		
		if isinstance(key, bool):
			if key:
				key = "_True"
			else:
				key = "_False"
		
		if isinstance(val, bool):
			if val:
				val = "_True"
			else:
				val = "_False"
		
		dict.__setitem__(self, key, val)
		dict.__setitem__(self, val, key)
	
	def __delitem__(self, key):
		dict.__delitem__(self, self[key])
		dict.__delitem__(self, key)
	
	def __getitem__(self, key):
		if key is None:
			key = "_None"
		
		if isinstance(key, bool):
			if key:
				key = "_True"
			else:
				key = "_False"
		
		val = dict.__getitem__(self, key)
		
		if val == "_None":
			return None
		elif val == "_True":
			return True
		elif val == "_False":
			return False
		else:
			return val
	
	def __contains__(self, key):
		if key is None:
			key = "_None"
		
		if isinstance(key, bool):
			if key:
				key = "_True"
			else:
				key = "_False"
		
		return dict.__contains__(self, key)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  nonelessdict.py
"""
Provides frozendict, a simple immutable dictionary.
"""
#
#  Copyright (c) 2020 Dominic Davis-Foster <dominic@davis-foster.co.uk>
#  Copyright (c) Jeremy Mayeres
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.
#


# stdlib
import operator
from collections import OrderedDict
from functools import reduce

# this package
from .base import MutableBase


class NonelessDict(MutableBase):
    """
    A wrapper around dict that will check if a value is None/empty/False,
    and not add the key in that case.
    Use the set_with_strict_none_check function to check only for None
    """
    
    dict_cls = dict

    def __init__(self, *args, **kwargs):
        if hasattr(self, "_dict"):
            raise TypeError(f"`{self.__class__}` can only be initialised once.")
    
        super().__init__(*args, **kwargs)

    def copy(self, **add_or_replace):
        return self.__class__(self, **add_or_replace)

    def __hash__(self):
        if self._hash is None:
            h = 0
            for key, value in self._dict.items():
                h ^= hash((key, value))
            self._hash = h
        return self._hash

    def set_with_strict_none_check(self, key, value):
        if value is not None:
            self._dict[key] = value


class NonelessOrderedDict(MutableBase):
    """
    A wrapper around OrderedDict that will check if a value is None/empty/False,
    and not add the key in that case.
    Use the set_with_strict_none_check function to check only for None
    """

    dict_cls = OrderedDict

    def __init__(self, *args, **kwargs):
        if hasattr(self, "_dict"):
            raise TypeError(f"`{self.__class__}` can only be initialised once.")

        super().__init__(*args, **kwargs)

    def copy(self, *args, **kwargs):
        new_dict = self._dict.copy()

        if args or kwargs:
            new_dict.update(OrderedDict(*args, **kwargs))

        return self.__class__(new_dict)

    def __hash__(self):
        if self._hash is None:
            self._hash = reduce(operator.xor, map(hash, self.items()), 0)

        return self._hash

    def set_with_strict_none_check(self, key, value):
        if value is not None:
            self._dict[key] = value



#!/usr/bin/env python
"""Setup script"""

from __pkginfo__ import (
	author, author_email, install_requires,
	license, long_description, classifiers,
	modname, short_desc, VERSION, web,
	project_urls, keywords,
	)

from setuptools import setup, find_packages

setup(
		author=author,
		author_email=author_email,
		classifiers=classifiers,
		description=short_desc,
		entry_points=None,
		install_requires=install_requires,
		license=license,
		long_description=long_description,
		name=modname,
		packages=find_packages(exclude=("tests",)),
		py_modules=None,
		url=web,
		project_urls=project_urls,
		version=VERSION,
		python_requires=">=3.6",
		keywords=keywords,
		)
